uniform float min_lod;
uniform sampler2D img_tex;
uniform sampler2D deg_tex;

void main(void)
{
  // values to convert rgb to grayscale from:
  // Foley et al, "Computer Graphics: Principles and Practice"
  vec4 lum = vec4( 0.299, 0.587, 0.114, 1.0 );

  // fetch sample from degradation texture
  vec4 deg = texture2D( deg_tex, gl_TexCoord[1].st );

  // the alpha channel of the degradation sample contains a lod bias.
  // this bias is in the range [0-1] where 1 is the highest
  // level of detail and 0 is the lowest.  we need to map
  // the alpha channel from [0-1] to [1-0].
  float lod = 1.0 - deg.w;

  // fetch sample from image texture, selecting lod obtained from
  // degradation sample, scaled from [1-0] to [1-min_lod] where min_lod is 
  // the coarsest mipmap level
  // (note: can't use texture2DLod() - that's only available in vertex prog)
  vec4 rgb = texture2D( img_tex, gl_TexCoord[0].st, lod * min_lod );

  // 1. scale the rgb channels of the image sample by the rgb channels of
  //    the degradation map (what's in the vec4())
  // 2. scale the rgb channels of the image sample by the 1 - the rgb
  //    channels of the degradation map (what's in the second arg of dot())
  // 3. return the rgb and lum samples
  gl_FragColor =
    vec4(rgb.xyz * deg.xyz,rgb.w) + dot(lum.xyz, rgb.xyz * (1.0 - deg.xyz));
}
