/****************************************************************************
** $Id: qt/gltexobj.cpp   3.1.2   edited Nov 8 2002 $
**
** Copyright (C) 1992-2002 Trolltech AS.  All rights reserved.
**
** This file is part of an example program for Qt.  This example
** program may be used, distributed and modified without limitation.
**
*****************************************************************************/

/****************************************************************************
**
** This is a simple QGLWidget demonstrating the use of QImages for textures.
**
** Much of the GL code is inspired by the 'spectex' and 'texcyl' 
** public domain demo programs by Brian Paul.
**
****************************************************************************/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>

#ifdef ANM_OSX
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <OpenGL/glext.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#endif

#include <qimage.h>
#include <qtimer.h>
#include <qevent.h>
#include <qfiledialog.h>
#include <qstring.h>

#include "gltexobj.h"

#include "utils.h"

// the ARB extension names should disappear once they are finally
// adopted in GL2.0 - the GL_GLEXT_PROTOTYPES comes from NVidia's
// glext.h header
#ifdef  GL_GLEXT_PROTOTYPES
#define GL_OBJECT_COMPILE_STATUS GL_OBJECT_COMPILE_STATUS_ARB
#define GL_OBJECT_LINK_STATUS GL_OBJECT_LINK_STATUS_ARB
#define GL_OBJECT_INFO_LOG_LENGTH GL_OBJECT_INFO_LOG_LENGTH_ARB

#define GLhandle GLhandleARB
#define GLchar GLcharARB

#define glUseProgramObject glUseProgramObjectARB
#define glActiveTexture glActiveTextureARB
#define glCreateShaderObject glCreateShaderObjectARB
#define glShaderSource glShaderSourceARB
#define glCompileShader glCompileShaderARB
#define glGetObjectParameteriv glGetObjectParameterivARB
#define glCreateProgramObject glCreateProgramObjectARB
#define glAttachObject glAttachObjectARB
#define glUniform3fv glUniform3fvARB
#define glGetUniformLocation glGetUniformLocationARB
#define glGetInfoLog glGetInfoLogARB
#define glDeleteObject glDeleteObjectARB
#define glLinkProgram glLinkProgramARB
#endif

#define printOpenGLError() printOglError(__FILE__, __LINE__)

// globals holding the GL context that we're storing our shaders in as well
// as handles to the vertex and fragment program used in this demo
GLhandle        svo;
GLhandle        sfo;
GLhandle        spo;

const int textures=2;
ulong texID[] = {GL_TEXTURE0,GL_TEXTURE1};

extern "C" {

void    printInfoLog(GLhandle);
int     printOglError(char *file, int line);
GLint   getUniLoc(GLhandle program, const GLchar *name);
char    *textFileRead(char *fn);

}

//const int redrawWait = 50;
const int redrawWait = 0;

/*!
  Create a GLTexobj widget
*/

GLTexobj::GLTexobj( QWidget* parent, const char* name ) :
	QGLWidget( parent, name ), \
	dx(0.0), \
	dy(0.0), \
	fps(0.0), \
	fps_samples(0.0), \
	fps_avg(0.0), \
	lod(4)
{
	uchar	red,green,blue,alpha=0xff;
	float	g,sigma=200.0;
	ofstream ofs;

// FOR mouse CONTROL, comment these 3 lines out
//GLTexobj::timer = new QTimer( this );
//connect( GLTexobj::timer, SIGNAL(timeout()), SLOT(update()) );
//GLTexobj::timer->start( redrawWait, FALSE );

  GLTexobj::img = new Ppm(512,512,0);

// DEBUG: output image we just created
//ofs.open("image.ppm");
//ofs << GLTexobj::img;
//ofs.close();

  GLTexobj::mask = new Ppm(GLTexobj::img->rows(),GLTexobj::img->cols(),0);
  for(int i=0; i<GLTexobj::mask->rows(); ++i) {
    for(int j=0; j<GLTexobj::mask->cols(); ++j) {
      g = gauss(i-GLTexobj::mask->rows()/2,j-GLTexobj::mask->cols()/2,sigma);
      red = green = blue = alpha = static_cast<uchar>(g * 255.0);
#ifdef ANM_OSX
      // store in RGBA format
        (*GLTexobj::mask)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#else
      // store in ABGR format (old way; for MESA)
        (*GLTexobj::mask)(i,j) = ((alpha<<24) | (blue<<16) | (green<<8) | red);
//      (*GLTexobj::mask)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#endif
    }
  }
// DEBUG: output mask we just created
  ofs.open("mask.ppm");
  ofs << GLTexobj::mask;
  ofs.close();
}


/*!
  Release allocated resources
*/

GLTexobj::~GLTexobj()
{
  makeCurrent();
}


/*!
  Set up the OpenGL rendering state, and define display list
*/

void GLTexobj::initializeGL()
{
	const GLubyte*	v = glGetString(GL_VENDOR);

  std::cerr << "GL: " << glGetString(GL_VERSION) << std::endl;

  if( (v[0] == 'N') && (v[1] == 'V') )		GLTexobj::vendor = NVIDIA;
  else if( (v[0] == 'A') && (v[1] == 'T') )	GLTexobj::vendor = ATI;
  else						GLTexobj::vendor = NVIDIA;

  std::cerr << "Loading shaders" << std::endl;
  loadShaders();
  std::cerr << "Loaded shaders" << std::endl;

  glGenTextures(textures,GLTexobj::texNames);

  imageTexture();
  imageMask();

  // specify clear color; (0.3,0.4,0.6,1.0) is a nice "slate blue"
  glClearColor( 0.0, 0.0, 0.0, 0.0 );

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
}


extern "C" {

GLint getUniLoc(GLhandle program, const GLchar *name)
{
        GLint   loc;
  
  loc = glGetUniformLocation(program,name);
  if(loc == -1)
    fprintf(stderr,"No such uniform named \"%s\"\n",name);
  printOpenGLError();
  return loc;
}

void printInfoLog(GLhandle obj)
{
        GLint   infologLength = 0;
        GLint   charsWritten = 0;
        GLchar  *infoLog;
  
  printOpenGLError();
  
  glGetObjectParameteriv(obj,GL_OBJECT_INFO_LOG_LENGTH,&infologLength);
  
  printOpenGLError();
  
  if(infologLength > 0) {
    infoLog = (GLchar *)malloc(infologLength);
    if(infoLog == NULL) {
      fprintf(stderr,"ERROR: Could not allocate InfoLog buffer\n");
      exit(1);
    }
    glGetInfoLog(obj,infologLength,&charsWritten,infoLog);
    if(charsWritten) fprintf(stderr,"InfoLog:\n%s\n\n",infoLog);
    free(infoLog);
  }
  printOpenGLError();
}

int printOglError(char *file, int line)
{
        GLenum  glErr;
        int     retCode = 0;
        char    *err;
  
  // returns 1 if an OpenGL error occurred, 0 otherwise.
  
  glErr = glGetError();
  while (glErr != GL_NO_ERROR) {
    err = (char*)gluErrorString(glErr);
    fprintf(stderr,"file: %s \n line: %d \n glError: %s\n",file,line,err);
    retCode = 1;
    glErr = glGetError();
  }
  return retCode;
}

}

/*!
  Set up the OpenGL view port, matrix mode, etc.
*/

void GLTexobj::resizeGL( int w, int h )
{
  glViewport(0,0,w,h);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0,w,0.0,h);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}


/*!
  Paint the texobj. The actual openGL commands for drawing the texobj are
  performed here.
*/

void GLTexobj::paintGL()
{
	float	lx,ly;

  // view
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0,width(),0.0,height());
  
  // object
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glDrawBuffer(GL_BACK);
  glClear(GL_COLOR_BUFFER_BIT);

  lx = (float)dx;
  ly = (float)dy;

  glDisable(GL_BLEND);
  glUseProgramObject(spo);

  // render
  glBegin(GL_QUADS);
  {
    glMultiTexCoord2f(texID[0],0,0);
    glMultiTexCoord2f(texID[1],      lx,      ly);
    glVertex2f(0,0);

    glMultiTexCoord2f(texID[0],1,0);
    glMultiTexCoord2f(texID[1],1.0 + lx,      ly);
    glVertex2f(width(),0);

    glMultiTexCoord2f(texID[0],1,1);
    glMultiTexCoord2f(texID[1],1.0 + lx,1.0 + ly);
    glVertex2f(width(),height());

    glMultiTexCoord2f(texID[0],0,1);
    glMultiTexCoord2f(texID[1],      lx,1.0 + ly);
    glVertex2f(0,height());
  }
  glEnd();

  swapBuffers();

  // clean up
  glUseProgramObject(0);
}


void GLTexobj::mouseMoveEvent(QMouseEvent* e)
{
//std::cerr << "(" << e->x() << "," << e->y() << ")" << std::endl;
  GLTexobj::dx = width()/2 - e->x();
  GLTexobj::dy = e->y() - height()/2;
  GLTexobj::dx /= width();
  GLTexobj::dy /= height();

  e->accept();
  updateGL();
}

void GLTexobj::update()
{
//	static Timer	timer;

//std::cerr << "updating..." << std::endl;

// can do timing here...

}

void GLTexobj::imageLOD(int ilod)
{
  GLTexobj::lod = ilod;

  // update shader linkage
  glUseProgramObject(spo);
  glUniform1f(getUniLoc(spo,"min_lod"),GLTexobj::lod);
  glUseProgramObject(0);

  updateGL();
}

void GLTexobj::imageDeg(int smoothing)
{
  // GL_TEXTURE_MIN_FILTER options:
  //                                                       (Manhattan dist.)
  //                                   intra-map interp    inter-map interp
  //   GL_NEAREST_MIPMAP_NEAREST:             no                  no
  //   GL_LINEAR_MIPMAP_NEAREST              yes                  no
  //   GL_NEAREST_MIPMAP_LINEAR               no                 yes
  //   GL_LINEAR_MIPMAP_LINEAR               yes                 yes
  // GL_TEXTURE_MAG_FILTER options:
  //   GL_NEAREST                             no                 n/a 
  //   GL_LINEAR                             yes                 n/a
  glActiveTexture(texID[0]);
  switch(smoothing) {
    case 1:
      // no smoothing
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_NEAREST_MIPMAP_NEAREST);
      break;
    case 2:
      // intra-map
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_LINEAR_MIPMAP_NEAREST);
      break;
    case 3:
      // inter-map
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_NEAREST_MIPMAP_LINEAR);
      break;
    case 4:
    default:
      // both
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_LINEAR_MIPMAP_LINEAR);
      break;
  }
  updateGL();
}

void GLTexobj::imageOpen()
{
    QString	qfilename = QFileDialog::getOpenFileName( QString::null,
                                                          QString::null,
                                                          this );
  if(!qfilename.isEmpty()) {
    std::ifstream ifs(qfilename.ascii());
    if(!ifs)
      std::cerr << "Warning: can't open " << qfilename << endl;
    else {
      ifs >> GLTexobj::img;
      ifs.close();

      imageTexture();

//    I don't think we need to do this...
//    GLTexobj::mask->resize(GLTexobj::img->rows(),GLTexobj::img->cols());
//    imageMask();
    }
  }
}


void GLTexobj::imageSaveAs()
{
	Ppm	img(height(),width(),0);

  glReadBuffer(GL_FRONT);
  glReadPixels(0,0,img.cols(),img.rows(),GL_RGBA,GL_UNSIGNED_BYTE,img.data());
                        
	QString qfilename = QFileDialog::getSaveFileName( "./",
                                                          "*.ppm",
                                                          this,
                                                          "save file dialog",
                                                          "Save As....");
  if(!qfilename.isEmpty()) {
    std::ofstream ofs(qfilename.ascii());
    if(!ofs)
      cerr << "Warning: can't open " << qfilename << endl;
    else {
      ofs << img;
      ofs.close();
    }   
  }     
}


void GLTexobj::openColorDeg()
{
    QString     qfilename = QFileDialog::getOpenFileName( QString::null,
                                                          QString::null,
                                                          this );
    Ppm                 *maskppm;
    register uchar      blue,green,red,alpha;
  
  if(!qfilename.isEmpty()) {
    std::ifstream ifs(qfilename.ascii());
    if(!ifs)
      std::cerr << "Warning: can't open " << qfilename << endl;
    else {
  
      maskppm = new Ppm(512,512,(ushort)0);
  
      ifs >> maskppm;
      ifs.close();
  
      maskppm->resize(GLTexobj::mask->rows(),GLTexobj::mask->cols());
  
      for(int i=0; i<GLTexobj::mask->rows(); ++i) {
        for(int j=0; j<GLTexobj::mask->cols(); ++j) {
  
          red       = (uchar)LONG_R((*maskppm)(i,j));
          green     = (uchar)LONG_G((*maskppm)(i,j));
          blue      = (uchar)LONG_B((*maskppm)(i,j));
          alpha     = (uchar)LONG_A((*GLTexobj::mask)(i,j));
          
#ifdef ANM_OSX
          // store in RGBA format
          (*mask)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#else
          // store in ABGR format (old way; for MESA)
          (*mask)(i,j) = ((alpha<<24) | (blue<<16) | (green<<8) | red);
//        (*mask)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#endif
        }  
      }

//    imageTexture();
      imageMask();

      delete maskppm;

    }
  }
} 


void GLTexobj::openSpatDeg()
{
    QString     qfilename = QFileDialog::getOpenFileName( QString::null,
                                                          QString::null,
                                                          this );
    Pgm                 *maskpgm;
    register uchar      blue,green,red,alpha;
    
  if(!qfilename.isEmpty()) {
    std::ifstream ifs(qfilename.ascii());
    if(!ifs)
      std::cerr << "Warning: can't open " << qfilename << endl;
    else {

      maskpgm = new Pgm(512,512,(ushort)0);

      ifs >> maskpgm;
      ifs.close();

      maskpgm->resize(GLTexobj::mask->rows(),GLTexobj::mask->cols());
  
      for(int i=0; i<GLTexobj::mask->rows(); ++i) {
        for(int j=0; j<GLTexobj::mask->cols(); ++j) {

          red       = (uchar)LONG_R((*GLTexobj::mask)(i,j));
          green     = (uchar)LONG_G((*GLTexobj::mask)(i,j));
          blue      = (uchar)LONG_B((*GLTexobj::mask)(i,j));
          alpha     = (uchar)SHORT_L((*maskpgm)(i,j));

#ifdef ANM_OSX
          // store in RGBA format
          (*GLTexobj::mask)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#else
          // store in ABGR format (old way; for MESA)
          (*GLTexobj::mask)(i,j) = ((alpha<<24) | (blue<<16) | (green<<8) | red);
//        (*GLTexobj::mask)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#endif
        } 
      }

//    imageTexture();
      imageMask();

      delete maskpgm;

    }
  }
}



void GLTexobj::imageTexture()
{
	GLint	error;

  if(glIsTexture(GLTexobj::texNames[0])) {
    glDeleteTextures(1,&GLTexobj::texNames[0]);
    glGenTextures(1,&GLTexobj::texNames[0]);
  }

  // texture 0 is the 'image'
  glActiveTexture(texID[0]);
  glBindTexture(GL_TEXTURE_2D,GLTexobj::texNames[0]);
  // set up texture env
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  error = gluBuild2DMipmaps(GL_TEXTURE_2D,
               GL_RGBA,                     // internalFormat
               GLTexobj::img->cols(),GLTexobj::img->rows(),     // width,height
               GL_RGBA,                     // format
               GL_UNSIGNED_BYTE,            // type
               GLTexobj::img->data());                // texels
  if(error)	std::cerr << gluErrorString(error) << std::endl;
  else		std::cerr << "gluBuild2DMipmaps succeeded." << std::endl;

}

void GLTexobj::imageMask()
{
  if(glIsTexture(GLTexobj::texNames[1])) {
    glDeleteTextures(1,&GLTexobj::texNames[1]);
    glGenTextures(1,&GLTexobj::texNames[1]);
  }

  // texture 1 is the mask
  // bind the alpha map (and create)
  glActiveTexture(texID[1]);
  glBindTexture(GL_TEXTURE_2D,GLTexobj::texNames[1]);
  // set up texture env
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glTexImage2D(GL_TEXTURE_2D,     // target
               0,                           // level
               GL_RGBA,                     // internalFormat
               GLTexobj::mask->cols(),GLTexobj::mask->rows(),   // width,height
               0,                           // border
               GL_RGBA,                     // format
               GL_UNSIGNED_BYTE,            // type
               GLTexobj::mask->data());               // texels
  // tell Cg which texture handle should be associated with the sampler2D
}

int GLTexobj::loadShaders()
{
        GLchar  *fs,*vs;
        GLint   vcompiled,fcompiled,linked;

  // create vertex shader object
  svo = glCreateShaderObject(GL_VERTEX_SHADER);

  // load source code strings into shader
  vs = textFileRead("vert.glsl");
  glShaderSource(svo,1,(const GLchar** )&vs,NULL);
  free(vs);

  // compile vertex shader and print compiler log file
  glCompileShader(svo);
  printOpenGLError();
  glGetObjectParameteriv(svo,GL_OBJECT_COMPILE_STATUS,&vcompiled);
  printInfoLog(svo);

  // check compile status
  if(!vcompiled) return 0;
  
  // create fragment shader object
  sfo = glCreateShaderObject(GL_FRAGMENT_SHADER);
  fs = textFileRead("frag.glsl");
  glShaderSource(sfo,1,(const GLchar** )&fs,NULL);
  free(fs);
  
  // compile fragment shader and print compiler log file
  glCompileShader(sfo);
  printOpenGLError();
  glGetObjectParameteriv(sfo,GL_OBJECT_COMPILE_STATUS,&fcompiled);
  printInfoLog(sfo);
  
  // check compile status
  if(!fcompiled) return 0;
  
  // create program object and attach the two compiled shaders
  spo = glCreateProgramObject();
  glAttachObject(spo,svo);
  glAttachObject(spo,sfo);
  
  // link the program object and print linker log file
  glLinkProgram(spo);
  printOpenGLError();
  glGetObjectParameteriv(spo,GL_OBJECT_LINK_STATUS,&linked);
  printInfoLog(spo);
  
  // check link status
  if(!linked) return 0;
  
  // clean up - detach shader objects following attachment to program object -
  // this will flag them for deletion as soon as they are no longer
  // referenced, i.e., when program object is deleted
  glDeleteObject(svo);
  glDeleteObject(sfo);
  
  // install program as part of current state
  glUseProgramObject(spo);
  
  // set up initial uniform values here
  // glUniform3f(getUniLoc(spo,"<uniform_var>"),<val1>,<val2>,<val3>);
  // glUniform3fv(getUniLoc(spo,"<uniform_var>"),<count>,<ptr>);
  glUniform1f(getUniLoc(spo,"min_lod"),lod);
  glUniform1i(getUniLoc(spo,"img_tex"),0);
  glUniform1i(getUniLoc(spo,"deg_tex"),1);
  
  // turn off all shaders for the time being
  glUseProgramObject(0);
  
  return 1;
}

extern "C" {

char *textFileRead(char *fn)
{
        FILE    *fp;
        char    *content = NULL;

        int     f,count;
  
  f = open(fn, O_RDONLY);
  count = lseek(f, 0, SEEK_END);
  close(f);
  
  if(fn != NULL) {
    fp = fopen(fn,"r");
    if(fp != NULL) {
      if(count > 0) {
        content = (char *)malloc(sizeof(char) * (count+1));
        count = fread(content,sizeof(char),count,fp);
        content[count] = '\0';
      }
      fclose(fp);
    }
  }
  return content;
}

int GLTexobj::textFileWrite(char *fn, char *s)
{
        FILE    *fp;
        int     status = 0;
  
  if(fn != NULL) {
    fp = fopen(fn,"w");
    if(fp != NULL) {
      if(fwrite(s,sizeof(char),strlen(s),fp) == strlen(s))
        status = 1;
      fclose(fp);
    }
  }
  return status;
}

}
