/****************************************************************************
** $Id: qt/glwinobj.cpp   3.1.2   edited Nov 8 2002 $
**
** Copyright (C) 1992-2000 Trolltech AS.  All rights reserved.
**
** This file is part of an example program for Qt.  This example
** program may be used, distributed and modified without limitation.
**
*****************************************************************************/

#include <iostream>
#include <qpushbutton.h>
#include <qslider.h>
#include <qlayout.h>
#include <qframe.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qapplication.h>
#include <qkeycode.h>

using namespace std;

#include <mlib/types.h>
#include <mlib/array.h>
#include <mlib/matrix.h>
#include <mlib/vector.h>
#include <mlib/quaternion.h>

#include <imglib/lum.h>
#include <imglib/rgba.h>
#include <imglib/smpte.h>
#include <imglib/pnm.h>
#include <imglib/ppm.h>

#include "glwinobj.h"
#include "gltexobj.h"


GLObjectWindow::GLObjectWindow( QWidget* parent, const char* name ) :
	QWidget( parent, name )
{
    // Create an OpenGL widget: (doubleBuffer | rgba | depth) set globally
    GLTexobj* c = new GLTexobj(this,"glarea");
    std::cout << "doubleBuffer: "	<< c->format().doubleBuffer()	<< " "
              << "rgba: "		<< c->format().rgba()		<< " "
              << "depth: "		<< c->format().depth()		<< " "
              << std::endl;
//  c->setMouseTracking(true);

    // create the file menu
    QPopupMenu *file = new QPopupMenu(this);
    file->insertItem("Open...",c,SLOT(imageOpen()),CTRL+Key_O);
    file->insertItem("Save As...",c,SLOT(imageSaveAs()),CTRL+Key_S);
    file->insertItem("Quit",qApp,SLOT(quit()),CTRL+Key_Q);

    // create degradation open subsubmenu
    QPopupMenu *iopen = new QPopupMenu(this);
    iopen->setAccel(0,iopen->insertItem("Color Image...",c,SLOT(openColorDeg()),-1,1));
    iopen->setAccel(0,iopen->insertItem("Spatial Image...",c,SLOT(openSpatDeg()),-1,2));

    // create degradation smoothing options subsubmenu
    QPopupMenu *ideg = new QPopupMenu(this);
    ideg->setAccel(0,ideg->insertItem("None",c,SLOT(imageDeg(int)),-1,1));
    ideg->setAccel(0,ideg->insertItem("Intra-map",c,SLOT(imageDeg(int)),-1,2));
    ideg->setAccel(0,ideg->insertItem("Inter-map",c,SLOT(imageDeg(int)),-1,3));
    ideg->setAccel(0,ideg->insertItem("Both",c,SLOT(imageDeg(int)),-1,4));
   
    // create degradation LOD options subsubmenu
    QPopupMenu *ilod = new QPopupMenu(this);
    ilod->setAccel(0,ilod->insertItem("1",c,SLOT(imageLOD(int)),-1,1));
    ilod->setAccel(0,ilod->insertItem("2",c,SLOT(imageLOD(int)),-1,2));
    ilod->setAccel(0,ilod->insertItem("3",c,SLOT(imageLOD(int)),-1,3));
    ilod->setAccel(0,ilod->insertItem("4",c,SLOT(imageLOD(int)),-1,4));
   
    QPopupMenu *degradation = new QPopupMenu(this);
    degradation->insertItem("Open",iopen);
    degradation->insertItem("Smoothing",ideg);
    degradation->insertItem("LOD",ilod);

    // create a menu bar
    QMenuBar *m = new QMenuBar(this);
    m->setSeparator(QMenuBar::InWindowsStyle);
    m->insertItem("File",file);
    m->insertItem("Degradation",degradation);

    // Top level layout (with 0,0 border)
    QHBoxLayout* hlayout = new QHBoxLayout( this, 0, 0, "hlayout");
    hlayout->setMenuBar(m);
    hlayout->addWidget( c, 1 );
}
